
function ArrowTree(/*String*/widgetId, /*Boolean*/folderOnly) { // constructor

	this.tree = dojo.widget.byId(widgetId);

	if (!this.tree) {
		return null;
	}

	// named constants
	this.NONE = 0;
	this.PARTIAL = 1;
	this.ALL = 2;

	this.UNDEFINED = 0;
	this.UNCHECKED = 1;
	this.ALLTOPARTIALCHECKED = 2;

	this.SCROLLDELTA = 5;

	this.numOfNode = Math.floor(dojo.html.getViewport().height / 20);

	this.folderOnly = folderOnly || false;

	this.checkboxStateChangedListener = [];

	/*------------------------------------------------------------------------*/

	this.reset();

	this.topicOfExpand = this.tree.widgetId + "/expand";
	this.topicOfCollapse = this.tree.widgetId + "/collapse";
	this.topicOfcreateDOMNode = this.tree.widgetId + "/createDOMNode";
	this.topicOfTitleClick = this.tree.widgetId + "/titleClick"

	dojo.event.topic.subscribe(this.topicOfExpand, this, "treeOnExpanded");
	dojo.event.topic.subscribe(this.topicOfCollapse, this, "treeOnCollapsed");
	dojo.event.topic.subscribe(this.topicOfcreateDOMNode, this, "insertCheckbox");
	dojo.event.topic.subscribe(this.topicOfTitleClick, this, "onTitleClicked");
}

ArrowTree.prototype.destroy = function() {
	this.reset();

	dojo.event.topic.unsubscribe(this.topicOfExpand, this, "treeOnExpanded");
	dojo.event.topic.unsubscribe(this.topicOfCollapse, this, "treeOnCollapsed");
	dojo.event.topic.unsubscribe(this.topicOfcreateDOMNode, this, "insertCheckbox");
	dojo.event.topic.unsubscribe(this.topicOfTitleClick, this, "onTitleClicked");
}

/*----------------------------------------------------------------------------*/

ArrowTree.prototype.treeOnExpanded = function(message) {

	var node = message.source;

	node.attrs.isExpanded = true;

	if (node.attrs.children !== null) { // no IO
		node.lock();

		dojo.lang.setTimeout(this,
							 function() {
								this.showTree(node, node.attrs);
								node.unlock();
							 },
							 200);
	}
	else {
		this.ioFunction(node);
	}
}


ArrowTree.prototype.open = function(/*String*/rootTitle) { // @public

	if (dojo.lang.isUndefined(rootTitle)) {
		return;
	}

	this.reset();

	var title = rootTitle;
	this.setChildren(this.tree, [{name: rootTitle, isFolder: true}]);

	var rootNode = this.tree.children[0];
	rootNode.expand();
}

ArrowTree.prototype.doRootIO = function(/*TreeNode*/node) {
	var tn = node;

	tn.lock(); // show loading icon

	var handler = function(res) {
		this.setChildren(tn, this.rootFormatter(res));

		tn.unlock();

		this.ioFunction = this.doChildrenIO;
	};

	var errorHandler = function() {
		tn.unlock();
	};

	this.rootIO(dojo.lang.hitch(this, handler),
				dojo.lang.hitch(this, errorHandler));
}


// doChildrenIO() should not be called directly
ArrowTree.prototype.doChildrenIO = function(/*TreeNode*/node) {
	var tn = node;

	tn.lock(); // show loading icon

	var handler = function(res) {
		this.setChildren(tn, this.childrenFormatter(res));

		tn.unlock();
	};

	var errorHandler = function() {
		tn.unlock();
	};

	this.childrenIO(dojo.lang.hitch(this, handler),
					dojo.lang.hitch(this, errorHandler), this.getPathComponents(tn));
}


ArrowTree.prototype.treeOnCollapsed = function(message) {

	var node = message.source;

	while (node.children.length > 0) {
		var n = node.children[0];

		this.destroyTreeNode(n);
	}
}

// clean all child widgets and model data except Tree widget
ArrowTree.prototype.reset = function() {
	while (this.tree.children.length > 0) {
		var n = this.tree.children[0];

		this.destroyTreeNode(n);
	}

	this.tree.attrs = {};
	this.folder = null;
	this.ioFunction = this.doRootIO;
}


ArrowTree.prototype.destroyTreeNode = function(/*TreeNode*/node) {

	while (node.children.length > 0) {
		var n = node.children[0];

		this.destroyTreeNode(n);
	}

	if (!dojo.lang.isUndefined(node.checkbox)) {

		dojo.event.disconnect(node.checkbox, "onClick", node.stateChangedCallback);

		node.checkbox.destroy();
	}

	node.attrs = null;

	node.parent.removeChild(node);

	node.destroy();
}


ArrowTree.prototype.setChildren = function(/*TreeNode*/tn, /*Array*/list) {

	tn.attrs.children = [];

	for (var i in list) {
		var e = list[i]

		tn.attrs.children.push({
			name: e.name,
			isFolder: e.isFolder,
			isExpanded: false,
			children: null,
			select: (tn.attrs.select == this.ALL && !this.folderOnly ? this.ALL : this.NONE)
		});
	}

	tn.attrs.scroll = (list.length > this.numOfNode);
	tn.attrs.first = 0; // index to tn.attrs.children
	tn.attrs.last = (list.length > this.numOfNode ? this.numOfNode - 1 : list.length - 1); // index to tn.attrs.children

	this.showTree(tn, tn.attrs);
}


ArrowTree.prototype.setIOHandler = function(/*Function*/rootIO, /*Function*/rootFormatter, /*Function*/childrenIO, /*Function*/childrenFormatter) {
	if (rootIO && rootFormatter && childrenIO && childrenFormatter) {
		this.rootIO = rootIO;
		this.rootFormatter = rootFormatter;
		this.childrenIO = childrenIO;
		this.childrenFormatter = childrenFormatter;
	}
}


ArrowTree.prototype.showTree = function(/*TreeNode*/tn, /*Object*/attrs) {

	for (var i = attrs.first, end = attrs.last + 1; i < end; i++) {

		var a = attrs.children[i];

		var node = this.showTreeNode(tn, a);
	}

	if (attrs.scroll) {
		var upArrowNode = this.showUpArrowTreeNode(tn);
		var downArrowNode = this.showDownArrowTreeNode(tn);
	}
}

ArrowTree.prototype.showTreeNode = function(/*TreeNode*/tn, /*Object*/attrs, /*Number*/index) {

	var node = this.createTreeNode(attrs.name.replace(/\s/g,"&nbsp;"), attrs.name, attrs.isFolder, attrs.isExpanded);

	node.attrs = attrs;

	tn.doAddChild(node, index);

	return node;
}

ArrowTree.prototype.showUpArrowTreeNode = function(/*TreeNode*/tn) {

	var node = this.createTreeNode("More ... ", null, false, false);

	// the up arrow icon is defined here
	var img = document.createElement('img');

	img.setAttribute("src", "../res/image/up0_on.gif");

	node.titleNode.appendChild(img);

	node.scrollUp = true;

	return this.addNoCheckboxTreeNode(tn, node, 0);
}

ArrowTree.prototype.showDownArrowTreeNode = function(/*TreeNode*/tn) {

	var node = this.createTreeNode("More ... ", null, false, false);

	// the down arrow icon is defined here
	var img = document.createElement("img");

	img.setAttribute("src", "../res/image/down0_on.gif");

	node.titleNode.appendChild(img);

	node.scrollDown = true;

	return this.addNoCheckboxTreeNode(tn, node);
}

ArrowTree.prototype.addNoCheckboxTreeNode = function(/*TreeNode*/parent, /*TreeNode*/node, /*Number*/index) {

	dojo.event.topic.unsubscribe(this.topicOfcreateDOMNode, this, "insertCheckbox");

	parent.doAddChild(node, index);

	dojo.event.topic.subscribe(this.topicOfcreateDOMNode, this, "insertCheckbox");

	return node;
}


ArrowTree.prototype.createTreeNode = function(/*String*/title, /*String*/objectId, /*Boolean*/isFolder, /*Boolean*/isExpanded) {

	var node = dojo.widget.createWidget("TreeNode", {
		title: title,
		objectId: objectId,
		isFolder: isFolder,
		expandLevel: (isExpanded ? 1 : 0)
	});

	return node;
}

/*----------------------------------------------------------------------------*/

ArrowTree.prototype.insertCheckbox = function(message) {

	var node = message.source;

	var checked = false;
	var disabled = false;

	if (node.attrs) {
		checked = (node.attrs.select === this.ALL ||
				   node.attrs.select === this.PARTIAL);

		disabled = (node.attrs.select === this.PARTIAL);
	}
	else if (node.parent.attrs) {
		checked = (node.parent.attrs.select === this.ALL)
	}

	var cb = dojo.widget.createWidget("Checkbox", {
		checked: checked,
		disabled: disabled
	});

	node.checkbox = cb;

	node.stateChangedCallback = dojo.lang.hitch(this, function() {
		this.onCheckboxStateChanged(node);
	});

	dojo.event.connect(cb, "onClick", node.stateChangedCallback);

	dojo.dom.insertBefore(cb.domNode, node.titleNode);
}

ArrowTree.prototype.addCheckboxStateChangedListener = function(/*Function*/func) {
	this.checkboxStateChangedListener.push(func);
}

ArrowTree.prototype.onCheckboxStateChanged = function(/*TreeNode*/tn) {

	if (this.folderOnly) {
		this.updateFolderSelection(tn);
	}
	else {
		this.updateFileSelection(tn);
	}


	for (var i in this.checkboxStateChangedListener) {
		var l = this.checkboxStateChangedListener[i];
		l();
	}
}


ArrowTree.prototype.updateFileSelection = function(/*TreeNode*/tn) {
	var checked = tn.checkbox.checked;

	this.changeChildrenState(tn.attrs, (checked ? this.ALL : this.NONE));

	var transition = this.changeParentState(tn.parent);

	this.updateChildrenState(tn);

	this.updateParentState(tn.parent);
}

// should execute in folderOnly mode
ArrowTree.prototype.updateFolderSelection = function(/*TreeNode*/tn) {
	var checked = tn.checkbox.checked;

	if (this.folder) { // clean previous select
		this.changeChildrenState(this.folder.attrs, this.NONE);
		this.changeParentStateFixed(this.folder.parent, this.NONE);
		this.updateChildrenState(this.folder);
		this.updateParentState(this.folder.parent);
	}

	this.folder = tn;

	tn.attrs.select = (checked ? this.ALL : this.NONE);
	this.changeParentStateFixed(tn.parent, (checked ? this.PARTIAL : this.NONE));

	this.updateChildrenState(tn);
	this.updateParentState(tn.parent);
}

ArrowTree.prototype.changeChildrenState = function(/*Object*/attrs, /*Number*/select) {

	attrs.select = select;

	for (var i in attrs.children) {
		var a = attrs.children[i];

		this.changeChildrenState(a, select);
	}
}

ArrowTree.prototype.changeParentStateFixed = function(/*TreeNode*/tn, /*Number*/select) {
	if (tn != this.tree) {
		var attrs = tn.attrs.select = select;

		this.changeParentStateFixed(tn.parent, select);
	}
}

ArrowTree.prototype.changeParentState = function(/*TreeNode*/tn) {

	var transition = this.UNDEFINED;

	if (tn != this.tree) { // tn is a tree node

		var oriselect = tn.attrs.select;

		var attrs = tn.attrs;

		var all = 0;
		var none = 0;

		for (var i in attrs.children) {
			var a = attrs.children[i];

			if (a.select === this.ALL) {
				all++;
			}
			else if (a.select === this.NONE) {
				none++;
			}
		}

		if (all === attrs.children.length) {
			attrs.select = this.ALL;
		}
		else if (none === attrs.children.length) {
			attrs.select = this.NONE;
		}
		else {
			attrs.select = this.PARTIAL;
		}

		this.changeParentState(tn.parent);

		if ((oriselect == this.ALL && attrs.select == this.NONE) ||
			(oriselect == this.PARTIAL && attrs.select == this.NONE)) {

			transition = this.UNCHECKED;
		}
		else if ((oriselect == this.ALL && attrs.select == this.PARTIAL)) {
			transition = this.ALLTOPARTIALCHECKED;
		}
	}

	return transition;
}

ArrowTree.prototype.updateChildrenState = function(/*TreeNode*/tn) {

	var cb = tn.checkbox;
	var select = tn.attrs.select;

	cb.checked = (select === this.ALL);
	cb.disabled = false;
	cb._setInfo();

	for (var i in tn.children) {
		var n = tn.children[i];

		if (dojo.lang.isUndefined(n.scrollUp) &&
			dojo.lang.isUndefined(n.scrollDown)) {

			this.updateChildrenState(n);
		}
	}
}

ArrowTree.prototype.updateParentState = function(/*TreeNode*/tn) {
	if (tn != this.tree) {

		var cb = tn.checkbox;
		var select = tn.attrs.select;

		cb.checked = (select === this.ALL || select === this.PARTIAL);
		cb.disabled = (select === this.PARTIAL);
		cb._setInfo();

		this.updateParentState(tn.parent);
	}
}


ArrowTree.prototype.onTitleClicked = function(message) {

	var tn = message.source;

	if (!dojo.lang.isUndefined(tn.scrollUp)) {

		var parent = tn.parent;

		for (var i = 0; i < this.SCROLLDELTA; i++) {
			if (parent.attrs.first > 0) {

				this.destroyTreeNode(parent.children[parent.children.length - 2]);

				var index = parent.attrs.first - 1;
				var a = parent.attrs.children[index];

				var node = this.showTreeNode(parent, a, 1);

				parent.attrs.first--; // add
				parent.attrs.last--; // remove
			}
		}
	}

	if (!dojo.lang.isUndefined(tn.scrollDown)) {

		var parent = tn.parent;


		for (var i = 0; i < this.SCROLLDELTA; i++) {
			if (parent.attrs.last < parent.attrs.children.length - 1) {

				this.destroyTreeNode(parent.children[1]);

				var index = parent.attrs.last + 1;
				var a = parent.attrs.children[index];

				var node = this.showTreeNode(parent, a, parent.children.length - 1);

				parent.attrs.first++; // remove
				parent.attrs.last++; // add

			}
		}
	}

	return false;
}

ArrowTree.prototype.getPathComponents = function(/*TreeNode*/tn) {
	var comps = [];

	for (var n = tn; n !== this.tree; n = n.parent) {
		comps.unshift(n.objectId);
	}

	return comps;
}

ArrowTree.prototype.getFullpath = function(/*TreeNode*/tn) {
	var c = this.getPathComponents(tn);

	return c.join("/");
}


ArrowTree.prototype.getSelectList = function() {

	var list = [];

	for (var i in this.tree.attrs.children) {
		var a = this.tree.attrs.children[i];

		this._getSelectList(a.name, a, list);
	}

	return list;
}


ArrowTree.prototype._getSelectList = function(/*String*/dir, /*Object*/attrs, /*Array*/list) {

	if (attrs.select === this.ALL) {
		list.push(dir + (attrs.isFolder ? "/" : ""));
		return;
	}

	for (var i in attrs.children) {
		var a = attrs.children[i];

		this._getSelectList(dir + "/" + a.name, a, list);
	}
}

/*----------------------------------------------------------------------------*/

// FIXME move utf8_decode() to String built-in type
ArrowTree.prototype.utf8_decode = function(/*String*/utftext) {
	var string = "";
	var i = 0;
	var c = c1 = c2 = 0;

	while ( i < utftext.length ) {

		c = utftext.charCodeAt(i);

		if (c < 128) {
			string += String.fromCharCode(c);
			i++;
		}
		else if((c > 191) && (c < 224)) {
			c2 = utftext.charCodeAt(i+1);
			string += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
			i += 2;
		}
		else {
			c2 = utftext.charCodeAt(i+1);
			c3 = utftext.charCodeAt(i+2);
			string += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
			i += 3;
		}

	}

	return string;
}
